#!/usr/bin/env perl

#(c) by Vlad Harchev <hvv@hippo.ru>
# Distributed under GPL

sub usage
{
    print "usage: keysymcooked-filename semicooked-filename outtablename testletter\n";
    exit 0;
};

#these two hashes use language-lettername as a key, and KeySym value as value
%ksup = ();
%kslow = ();

#letters - upper- and lowercased
%lup = ();
%llow = ();

sub readkeysymcooked
{
    $nm = shift;
    open(F,"<$nm") || die("can't open $nm");
    @lines = <F>;
    close(F);        
    for $l (@lines) {
	my ($lang,$letter,$val) = split(/\s+/,$l);
	$lang = lc $lang;	
	$val = hex $val;
	if ( (uc $letter) eq $letter) {
	    $ksup{$lang . "-" . (lc $letter)} = $val;
	} else {
	    $kslow{$lang . "-" . (lc $letter)} = $val;
	};
    }
}

sub readsemicooked
{
    $nm = shift;
    open(F,"<$nm") || die("can't open $nm");
    @lines = <F>;
    close(F);        
    for $l (@lines) {
	my ($val,$lang,$case,@name) = split(/\s+/,$l);
	$lang = lc $lang;	
	$val = hex $val;
	$letter = lc (join("",@name));
	if ( (uc $case) eq "CAPITAL" ) {
	    $lup{$lang . "-" . $letter} = $val;
	} else {
	    $llow{$lang . "-" . $letter} = $val;
	};
    }    
};

sub getminmax
{
    my (@list) = @_;
    my($min,$max) = (@list[0],@list[0]);
    for $i (@list) {
	if ($min > $i) {
	    $min = $i;
	};
	if ($max < $i) {
	    $max = $i;
	};	    
    };
    ($min,$max);
}

sub getnonzeropos
{
    my (@list) = @_;
    my ($idx) = 0;
    loop: for $i (@list) {
	last loop if ($i != 0);
	++$idx;
    };    
    $idx;
};

sub getrangeinfo
{
    my (@list) = @_;
    my ($minidx,$maxidx) = (getnonzeropos(@list),
	    @list - 1 - getnonzeropos(reverse(@list)));    
    ($minidx,$maxidx);
};

sub gentable
{
    #these hold ranges of keysyms' values for each of cases that keysym hold
    my ($nm,$testletter) = @_;
    ($min,$max) = getminmax(values %ksup,values %kslow);
    my (@tbl) = (0) x ($max-$min +1 );
    my (@comment) = ("") x ($max-$min +1 );

    print "static char_type ${nm}_to_str[]={\n";
    
    for $case (0..1) {
	%ks = ($case ? %ksup : %kslow);
	%ls = ($case ? %lup : %llow);
	my ($caselabel) = ($case ? "capital " : "small ");
	
	for $ks (keys %ks) {
	    $ksval = $ks{$ks};    
	    if (defined($ls{$ks})) {
		@tbl[$ksval - $min] = $ls{$ks};
		@comment[$ksval - $min] = $caselabel. $ks;
	    };
	}
    };
    my ($minidx,$maxidx) = getrangeinfo(@tbl);
    for ($i=$minidx;$i<=$maxidx;++$i) {
        print "\t",(sprintf "0x%x", @tbl[$i]), ",\t/* ",@comment[$i]," */\n";
    }	
    print "\n};\n\n";
    
    die("test letter is not present in both cases!") 
	if  ( (!defined $llow{$testletter}) || (!defined $lup{$testletter}) );
    print "\nstatic locale_info ${nm}_info = \n{ " , "\"$nm\", ",
	($min+$minidx) , ", " , ($min+$maxidx) , " , ${nm}_to_str, ",
	sprintf("0x%x",$llow{$testletter}), ", ", sprintf("0x%x",$lup{$testletter}), " };\n\n";
};

#this is used only for generation of table for vncviewer for Win
sub genchartokstable
{
    my (@tbl) = (0) x 256;
    my (@comment) = ("") x 256;    
    for $case (0..1) {
	%ks = ($case ? %ksup : %kslow);
	%ls = ($case ? %lup : %llow);
	
	for $l (keys %ls) {
	    if (defined $ks{$l}) {
		my ($idx) = $ls{$l};
		@tbl[$idx] = $ks{$l};
		@comment[$idx] = ($case ? "capital "  : "small ") . $l;
	    };
	};
    };
    for $i (0..(@tbl-1) ) {
	if (@tbl[$i]>0) {
	    print (sprintf "	0x%x",@tbl[$i]);
	    print ",\t/* " . @comment[$i] . " */\n";
	} else {
	    print "	0,\n";
	};
    };
};

if (1) {
    (@ARGV !=4 ) && usage();
    ($keysymsname,$intblnm,$tblnm,$testletter) = @ARGV;
} else {
    ($keysymsname,$intblnm,$tblnm,$testletter) = 
	("keysyms.cooked","cp1251.semicooked","foo","cyrillic-a");
};

readkeysymcooked ($keysymsname);
readsemicooked($intblnm);
if (1) {
    gentable($tblnm,$testletter);
} else {
    #this branch is used only when generating tables for winvncviewer.
    #call it manually
    genchartokstable();
}

0;

