/*
    (c) by Vlad Harchev <hvv@hippo.ru>
    
    Distributed under GPL.

Changelog:
    1.0 Initial release.
    
*/

#include <dlfcn.h>  
#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#define XK_CYRILLIC
#include <X11/keysym.h>

#ifndef RTLD_NEXT
# define RTLD_NEXT      ((void *) -1l)
#endif

static int trace = 0;

#undef FB
#undef DBG

#define DBG(x) do { if (trace) printf x; } while (0)
#define FB(x)  do { DBG(("falling back: ")); DBG(x); DBG(("\n")); goto fallback; } while (0)
/*
 * returns a pointer to the implementation of 'funcName' in
 * the libc library.  If not found, terminates the program.
 */
static void *getLibraryFunction(const char *funcName)
{
	void *res;

	if ((res = dlsym(RTLD_NEXT, funcName)) == NULL) {
		fprintf(stderr, "dlsym %s error:%s\n", funcName,
			dlerror());
		_exit(1);
	}
	return res;
}

typedef char* (*XKeysymToString_t)(KeySym val);
static XKeysymToString_t real_XKeysymToString;

static int convert(int ks,char* out,int bytes_in_buffer);

typedef int (*_XLookupString_fn_t)(XKeyEvent *event_struct, char *buffer_return, int bytes_buffer,
     KeySym *keysym_return, void *status_in_out);
typedef KeySym (*XLookupKeysym_t)(XKeyEvent *key_event,int index);
typedef KeyCode (*XKeysymToKeycode_t)(Display *display,KeySym keysym);
typedef void (*XConvertCase_t)(KeySym keysym,KeySym* lower,KeySym* upper);


int XLookupString(XKeyEvent *event, char *buffer_return, int bytes_buffer,
 KeySym *keysym_return, void *status_in_out)
{
    static _XLookupString_fn_t real__XLookupString = NULL;
    static XLookupKeysym_t real_XLookupKeysym;
    static XKeysymToKeycode_t real_XKeysymToKeycode;
    static XConvertCase_t real_XConvertCase;
    
    static int ignore_vendor = 0;
    
    KeySym ks;
    int cnt;
    int need_upper;/*whether uppercased version is needed*/
    
    if (!real__XLookupString) {
	real__XLookupString = 	(_XLookupString_fn_t) 	getLibraryFunction("XLookupString");
	real_XLookupKeysym = 	(XLookupKeysym_t) 	getLibraryFunction("XLookupKeysym");
	real_XKeysymToKeycode = (XKeysymToKeycode_t) 	getLibraryFunction("XKeysymToKeycode");
	real_XConvertCase = 	(XConvertCase_t) 	getLibraryFunction("XConvertCase");
	real_XKeysymToString = (XKeysymToString_t)	getLibraryFunction("XKeysymToString");
	
	if (getenv("VNCKBEMU_TRACE"))
	    trace = 1;
	if (getenv("VNCKBEMU_ANY_XSERVER"))
	    ignore_vendor = 1;
    };
    if (!ignore_vendor) {
	char* vendorname = ServerVendor(event->display);
	if (! ((vendorname[0]=='A' || vendorname[1]=='T' || vendorname[2]=='&') ||
		(vendorname[0]=='R' || vendorname[1]=='e' || vendorname[2]=='a'))) 
	    goto fallback;
    };
    ks = (*real_XLookupKeysym)(event,0);
    if (ks==NoSymbol)
	FB(("XLookupKeysym can't say anything\n"));
    if (!buffer_return)
	FB(("empty buffer_return given\n"));
    {	
	/* determine the proper case of the letter 
	    (and whether the string should be returned at all).	 
	*/
	KeyCode keycode_orig = event->keycode;
	KeyCode kc_a = (*real_XKeysymToKeycode)(event->display,XK_a);
	if (!kc_a)
	    FB(("no key with keysym 'XK_a'!"));
	event->keycode = kc_a;
	cnt = (*real__XLookupString)(event, buffer_return, bytes_buffer,
	    keysym_return, 0 ? status_in_out : NULL );
	event->keycode = keycode_orig;
	if (!cnt) /*probably nothing should be returned*/
	    FB(("real XLookupString doesn't know what to return even for 'a'!"));
	if (cnt!=1)
	    FB(("real XLookupString returned %d chars",cnt));
	if (!(*buffer_return=='a' || *buffer_return=='A'))
	    /*something wrong or compose was active*/
	    FB(("XLookupString returned '%s' while expected 'a' or 'A'",buffer_return));
	need_upper = *buffer_return == 'A';
    }    
    if (keysym_return)
        *keysym_return = ks;
    {
	KeySym low,upper;
	(*real_XConvertCase)(ks,&low,&upper);
	if (cnt=convert(need_upper ? upper : low ,buffer_return, bytes_buffer))
	    return cnt;	    
    }
fallback:
    return (*real__XLookupString)(event, buffer_return, bytes_buffer,
	    keysym_return, status_in_out);
};
 
									       
typedef unsigned char char_type;

typedef struct 
{
    char* name; /* mostly for debugging */

    int minks,maxks;
    char_type* convtable;
    char_type sample_lower,sample_upper; 
	/*
	    if (tolower(sample_upper)==sample_lower)
	    then this instance 'locale_info' should be used for conversion.
	*/
} locale_info;

#include "tables.cP"

static char_type void_to_str = 'w';

/*this is fallback - it doesn't translate anything*/
static locale_info void_info = 
{ "unknown", (int)'w',(int)'w', &void_to_str, 'w', 'W' };

static locale_info* locale_infos[] = {
    &cp1251_info,
    &koi8ub_info,
    &iso5_info,
    &void_info,
    NULL
};

/*returns # of bytes taken by translation*/
static int convert(int ks,char* out,int bytes_in_buffer)
{
    static locale_info* xlation_info = NULL;
    if (!xlation_info) {
	int i;
	char* enc;
	DBG(("known encodings:"));
	for(i=0;locale_infos[i];++i) {	
	    DBG((" '%s'",locale_infos[i]->name));
	}
	DBG(("\n"));
	for(i=0;locale_infos[i];++i) {
	    if (tolower(locale_infos[i]->sample_upper)==locale_infos[i]->sample_lower) {
		xlation_info = locale_infos[i];
		break;
	    };
	};
	if (!xlation_info) /*can't happen*/
	    xlation_info = &void_info;
	if (enc=getenv("VNCKBEMU_FORCE_ENCODING")) {
	    DBG(("autodetected encoding '%s'\n",xlation_info->name));
	    DBG(("while searching for forced encoding '%s' found ",enc));
	    xlation_info = &void_info;	    
	    for(i=0;locale_infos[i];++i) {
		if (!strcasecmp(enc,locale_infos[i]->name)) {
		    xlation_info = locale_infos[i];
		    break;
		};
	    };	    
	    DBG(("'%s', using it\n",xlation_info->name));
	};
	if (getenv("VNCKBEMU_DISABLE")) {
	    xlation_info = &void_info;	
	};
	DBG(("using encoding %s\n",xlation_info->name));
    };
    DBG(("keysym is 0x%x\n",ks));
    if (ks < xlation_info->minks || ks > xlation_info->maxks || bytes_in_buffer<2)
	return 0;
    else {
	out[0] = xlation_info->convtable[ks - xlation_info->minks];
	out[1] = '\0';
	DBG(("converted to '%s'\n",out));
	return 1;
    };
};
